"""
diversipy is a collection of algorithms for sampling in hypercubes,
selecting diverse subsets, and measuring diversity.

For further documentation see the submodules.
"""

from diversipy.hycusampling import *
from diversipy.subset import *
from diversipy.indicator import *
from diversipy.distance import *


__version__ = "0.6"
